#ifdef USE_FLI
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "xdvi.h"

static char *strspc(char *s)
{
    while (*s) {
	switch (*s) {
	case ' ':
	case '\n':
	case '\t':
	    return s;
	default:
	}
	++s;
    }
    return NULL;
}


struct fontlib {
    char *fontlibname;
    int numsizes;
    struct fontsize *sizeblks;
    struct fontlib *next;
} *fontlibfirst = NULL;

struct fontsize {
    int numfiles;
    int roundedpksize;
    double pksize;
    unsigned char *list;
};

static int get_2(unsigned char *p)
{   return ((int)*p << 8) + (int)(p[1]);
}

static long get_4(unsigned char *p)
{   return ((long)*p << 24) + ((long)(p[1]) << 16) + 
	((long)(p[2]) << 8) + (long)(p[3]);
}

void *malloc_(int n)
{   void *p;
    if (!(p = malloc(n))) {
	fprintf(stderr, "out of memory\n");
	exit(1);
    }
    return p;
}

void read_font_library(char *filename)
{   FILE *fp;
    unsigned char hdr[14];
    struct fontlib *fl = NULL;
    int i, numsizes;

    if ((fp = fopen(filename, "r")) == NULL) {
	fprintf(stderr, "cannot open font library %s\n", filename);
	return;
    }
#ifdef DEBUG
    fprintf(stderr, "reading font library %s ", filename);
#endif
    if (fread(hdr, 1, 14, fp) != 14) {
	fclose(fp);
read_err:
	fprintf(stderr, "read error\n");
	if (fl) {
	    fontlibfirst = fl->next;
	    for (i = 0; i < numsizes; ++i)
		if (fl->sizeblks[i].list)
		    free(fl->sizeblks[i].list);
	    free(fl->sizeblks);
	    free(fl);
	}
	return;
    }
    if (memcmp(hdr, "FLIB\2", 6)) {
	fprintf(stderr, "not in Fontlib/2-format\n");
	fclose(fp);
	return;
    }
    numsizes = get_2(hdr+8);
#ifdef DEBUG
    fprintf(stderr, "(%d sizes) ", numsizes);
#endif
    if (numsizes < 0 || numsizes > 100)
	goto read_err;
    fl = malloc_(sizeof(struct fontlib));
    fl->fontlibname = malloc_(strlen(filename) + 1);
    strcpy(fl->fontlibname, filename);
    fl->next = fontlibfirst;
    fontlibfirst = fl;
    fl->numsizes = numsizes;
    fl->sizeblks = malloc_(numsizes * sizeof(struct fontsize));
    for (i = 0; i < numsizes; ++i)
	fl->sizeblks[i].list = NULL;	/* init, falls Fehler */
#ifdef DEBUG
    fprintf(stderr, "(%d files)\n", get_2(hdr+10));
#endif
    fseek(fp, (long)get_2(hdr+12), 1);	/* skip comment */

    for (i = 0; i < numsizes; ++i) {
	int numbytes;

	if (fread(hdr, 1, 8, fp) != 8)
	    goto read_err;
	fl->sizeblks[i].list = malloc_(numbytes = get_2(hdr));
	fl->sizeblks[i].numfiles = get_2(hdr+2);
	fl->sizeblks[i].roundedpksize = get_2(hdr+4);
	fl->sizeblks[i].pksize = (double)(fl->sizeblks[i].roundedpksize) +
			(double)(unsigned)get_2(hdr+6) / 65536.0;
	if (hdr[6] & 0x80)
	    ++fl->sizeblks[i].roundedpksize;
#ifdef DEBUG
	fprintf(stderr, "(%d fonts at size %f)\n", fl->sizeblks[i].numfiles,
	    fl->sizeblks[i].pksize);
#endif
        if (fread(fl->sizeblks[i].list, 1, numbytes, fp) != numbytes)
	    goto read_err;
    }
    fclose(fp);
    return;
}

void read_all_fontlibs(void)
{   char *flis;
    if ((flis = getenv("XDVIFLI"))) {
	char *s, *t;
	s = malloc_(strlen(flis)+1);
	strcpy(s, flis);
	flis = s;
	while ((t = strspc(s))) {
	    *t = '\0';
	    if (s != t)
	      read_font_library(s);
	    s = t+1;
	}
	if (*s)
	    read_font_library(s);
	free(flis);
    }
}

FILE *fli_open(const unsigned char *font, int size, char **name)
{   struct fontlib *fl;
    int i;

    for (fl = fontlibfirst; fl; fl = fl->next) {
	for (i = 0; i < fl->numsizes; ++i) {
	    if (fl->sizeblks[i].roundedpksize == size) {
		/* look for the font */
		unsigned char *s;
		int numfonts, j;
		numfonts = fl->sizeblks[i].numfiles;
		s = fl->sizeblks[i].list;
		for (j = 0; j < numfonts; ++j) {
		    int len;
		    len = s[8];
		    if (!memcmp(s+9, font, len)) {
			FILE *fp;
			fprintf(stderr, "found font %s, size %d in fontlib %s\n", font, size,
			    fl->fontlibname);
			if (name) {
			    *name = malloc_(strlen(fl->fontlibname)+1);
			    strcpy(*name, fl->fontlibname);
			}
			fp = xfopen(fl->fontlibname);
			if (!fp) {
			    fprintf(stderr, "error opening fontlib!!!!\n");
			    exit(1);
			}
			fseek(fp, get_4(s+4), 0);
			return fp;
		    }
		    s += 9 + len;
		}
		/* font not found */
	    }
	}
    }
    return NULL;
}
#endif
